#ifndef OSM2PGSQL_PROJECTION_HPP
#define OSM2PGSQL_PROJECTION_HPP

/**
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * This file is part of osm2pgsql (https://osm2pgsql.org/).
 *
 * Copyright (C) 2006-2026 by the osm2pgsql developer community.
 * For a full list of authors see the git log.
 */

constexpr int PROJ_LATLONG = 4326;
constexpr int PROJ_SPHERE_MERC = 3857;

#endif // OSM2PGSQL_PROJECTION_HPP
